# -*- coding: utf-8 -*-

from collections import OrderedDict

import tala.nl.gf.resource
from tala.ddd.ddd_py_compiler import DddPyCompiler, DddPyCompilerException, GrammarCompiler
from tala.ddd.parser import Parser
from tala.ddd.test.ddd_compiler_test_case import DddCompilerTestCase
from tala.model.ontology import Ontology
from tala.model.plan import Plan
from tala.model.plan_item import ForgetAllPlanItem
from tala.model.predicate import Predicate
from tala.model.proposition import PredicateProposition
from tala.model.sort import CustomSort, RealSort
from tala.model.individual import Individual
from tala.nl.gf.grammar_entry_types import Constants, Node


class MockGrammarCompiler(GrammarCompiler):
    def _warn(self, warning):
        self.actual_warning = warning


class DddPyCompilerTestCase(DddCompilerTestCase):
    def _given_compiled_ontology(self, domain_name="Domain", **kwargs):
        ontology_args = self._compile_ontology(**kwargs)
        self._ontology = Ontology(**ontology_args)
        self._parser = Parser(self.DDD_NAME, self._ontology, domain_name)

    def _when_compile_ontology(self, **kwargs):
        self._result = self._compile_ontology(**kwargs)

    def _compile_ontology(self, name="Ontology", sorts={}, predicates={}, individuals={}, actions=set()):
        class Ontology:
            pass

        Ontology.__name__ = name
        Ontology.sorts = sorts
        Ontology.predicates = predicates
        Ontology.individuals = individuals
        Ontology.actions = actions
        self._ontology_name = name
        return DddPyCompiler().compile_ontology(Ontology)

    def _when_compile_domain(self, **kwargs):
        class Domain:
            plans = []

        for key, value in kwargs.iteritems():
            setattr(Domain, key, value)

        self._result = DddPyCompiler().compile_domain(self.DDD_NAME, Domain, self._ontology, self._parser)

    def _when_compile_domain_with_plan(self, plan):
        self._when_compile_domain(plans=[{"goal": "perform(top)", "plan": plan}])

    def _when_compile_plan_with_attribute(self, key, value):
        self._when_compile_domain(plans=[{"goal": "perform(top)", "plan": [], key: value}])

    def _then_result_has_keys(self, expected_keys):
        self.assertItemsEqual(expected_keys, self._result.keys())

    def _then_result_has_field(self, expected_key, expected_value):
        self.assertEquals(expected_value, self._result[expected_key])

    def _then_result_has_plan(self, expected_plan):
        actual_plans = self._result["plans"]
        actual_plan = actual_plans[0]["plan"]
        self.assertEquals(expected_plan, actual_plan)

    def _then_result_has_plan_with_attribute(self, expected_key, expected_value):
        actual_plans = self._result["plans"]
        actual_plan = actual_plans[0]
        self.assertEquals(expected_value, actual_plan[expected_key])

    def _parse(self, string):
        return self._parser.parse(string)

    def _when_compile_grammar(self, *args):
        self._compile_grammar(*args)

    def _compile_grammar(self, string):
        self._grammar_compiler = MockGrammarCompiler()
        self._result = self._grammar_compiler.compile(string, self._ontology, self._service_interface)

    def _then_result_is(self, expected_result):
        self.assertEquals(expected_result, self._result)

    def _then_grammar_is(self, expected_grammar_children):
        expected_grammar_node = Node(Constants.GRAMMAR, {}, expected_grammar_children)

        self.assertEquals(
            expected_grammar_node, self._result,
            "\n%s\n!=\n%s\n\n(%r)" % (expected_grammar_node, self._result, self._result)
        )

    def _then_warning(self, expected_warning):
        self.assertEquals(expected_warning, self._grammar_compiler.actual_warning)

    def _predicate(self, *args, **kwargs):
        return Predicate(self._ontology_name, *args, **kwargs)

    def _individual(self, *args, **kwargs):
        return Individual(self._ontology_name, *args, **kwargs)


class OntologyCompilerTests(DddPyCompilerTestCase):
    def test_name(self):
        self._when_compile_ontology(name="MockupOntology")
        self._then_result_has_field("name", "MockupOntology")

    def test_custom_sort(self):
        self._when_compile_ontology(sorts={"city": {}})
        self._then_result_has_field("sorts", set([CustomSort(self._ontology_name, "city")]))

    def test_dynamic_sort(self):
        self._when_compile_ontology(sorts={"city": {"dynamic": True}})
        self._then_result_has_field("sorts", set([CustomSort(self._ontology_name, "city", dynamic=True)]))

    def test_predicate_of_custom_sort(self):
        self._when_compile_ontology(predicates={"dest_city": "city"}, sorts={"city": {}})
        self._then_result_has_field(
            "predicates", set([self._predicate("dest_city", CustomSort(self._ontology_name, "city"))])
        )

    def test_predicate_of_builtin_sort(self):
        self._when_compile_ontology(predicates={"price": "real"})
        self._then_result_has_field("predicates", set([self._predicate("price", RealSort())]))

    def test_predicate_of_dynamic_sort(self):
        self._when_compile_ontology(predicates={"dest_city": "city"}, sorts={"city": {"dynamic": True}})
        self._then_result_has_field(
            "predicates", set([self._predicate("dest_city", CustomSort(self._ontology_name, "city", dynamic=True))])
        )

    def test_feature_of(self):
        self._when_compile_ontology(
            sorts={"city": {}},
            predicates={
                "dest_city": "city",
                "dest_city_type": {
                    "sort": "city",
                    "feature_of": "dest_city"
                }
            }
        )
        self._then_result_has_field(
            "predicates",
            set([
                self._predicate("dest_city", sort=CustomSort(self._ontology_name, "city")),
                self._predicate(
                    "dest_city_type", sort=CustomSort(self._ontology_name, "city"), feature_of_name="dest_city"
                )
            ])
        )

    def test_individuals(self):
        self._when_compile_ontology(individuals={"paris": "city"}, sorts={"city": {}})
        self._then_result_has_field("individuals", {"paris": CustomSort(self._ontology_name, "city")})

    def test_actions(self):
        self._when_compile_ontology(actions=set(["buy"]))
        self._then_result_has_field("actions", set(["buy"]))


class PlanCompilationTests(DddPyCompilerTestCase):
    def test_plan_for_perform_goal(self):
        self._given_compiled_ontology()

        self._when_compile_domain(plans=[{"goal": "perform(top)", "plan": []}])

        self._then_result_has_field("plans", [{"goal": self._parse("perform(top)"), "plan": Plan([])}])

    def test_plan_for_resolve_goal(self):
        self._given_compiled_ontology(predicates={"price": "real"})

        self._when_compile_domain(plans=[{"goal": "resolve(?X.price(X))", "plan": []}])

        self._then_result_has_field("plans", [{"goal": self._parse("resolve(?X.price(X))"), "plan": Plan([])}])

    def test_exception_yielded_if_plans_is_not_list(self):
        self._given_compiled_ontology()
        with self.assertRaises(DddPyCompilerException):
            self._when_compile_domain(plans={})

    def test_exception_yielded_if_plan_not_list(self):
        self._given_compiled_ontology()
        with self.assertRaises(DddPyCompilerException):
            self._when_compile_domain(plans=[{"goal": "perform(top)", "plan": None}])

    def test_plan_for_action_yields_helpful_exception(self):
        self._given_compiled_ontology()

        with self.assertRaises(DddPyCompilerException) as cm:
            self._when_compile_domain(plans=[{"goal": "top", "plan": []}])

        self.assertTrue('"top" is not a goal. Perhaps you mean "perform(top)"?' in unicode(cm.exception))

    def test_plan_for_issue_yields_helpful_exception(self):
        self._given_compiled_ontology(predicates={"price": "real"})

        with self.assertRaises(DddPyCompilerException) as cm:
            self._when_compile_domain(plans=[{"goal": "?X.price(X)", "plan": []}])

        self.assertTrue(
            '"?X.price(X)" is not a goal. '
            'Perhaps you mean "resolve(?X.price(X))"?' in unicode(cm.exception)
        )

    def test_plan_stack_order(self):
        self._given_compiled_ontology(
            sorts={
                "how": {},
                "city": {},
            }, predicates={
                "means_of_transport": "how",
                "dest_city": "city",
            }
        )

        self._when_compile_domain_with_plan(["findout(?X.means_of_transport(X))", "findout(?X.dest_city(X))"])

        self._then_result_has_plan(
            Plan([self._parse("findout(?X.dest_city(X))"),
                  self._parse("findout(?X.means_of_transport(X))")])
        )

    def test_preferred(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("preferred", True)
        self._then_result_has_plan_with_attribute("preferred", True)

    def test_accommodate_without_feedback(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("accommodate_without_feedback", True)
        self._then_result_has_plan_with_attribute("accommodate_without_feedback", True)

    def test_dynamic_title(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("dynamic_title", True)
        self._then_result_has_plan_with_attribute("dynamic_title", True)

    def test_restart_on_completion(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("restart_on_completion", True)
        self._then_result_has_plan_with_attribute("restart_on_completion", True)

    def test_gui_context(self):
        self._given_compiled_ontology(predicates={"price": "real"})
        self._when_compile_plan_with_attribute("gui_context", ["price"])
        self._then_result_has_plan_with_attribute("gui_context", [self._ontology.get_predicate("price")])

    def test_postplan(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("postplan", ["forget_all"])
        self._then_result_has_plan_with_attribute("postplan", Plan([ForgetAllPlanItem()]))

    def test_postconds(self):
        self._given_compiled_ontology(
            predicates={"dest_city": "city"}, sorts={"city": {}}, individuals={"gothenburg": "city"}
        )
        self._when_compile_plan_with_attribute("postconds", ["dest_city(gothenburg)"])
        self._then_result_has_plan_with_attribute(
            "postconds", [
                PredicateProposition(
                    self._predicate("dest_city", CustomSort(self._ontology_name, "city")),
                    self._individual("gothenburg", CustomSort(self._ontology_name, "city"))
                )
            ]
        )

    def test_unsupported_key_in_plan_description_yields_exception(self):
        self._given_compiled_ontology()
        with self.assertRaises(DddPyCompilerException):
            self._when_compile_domain(plans=[{"goal": "perform(top)", "plan": [], "unsupported_key": True}])


class DomainCompilerTests(DddPyCompilerTestCase):
    def test_name(self):
        self._given_compiled_ontology()
        self._when_compile_domain()
        self._then_result_has_field("name", "Domain")

    def test_dependency(self):
        self._given_compiled_ontology(
            sorts={
                "city": {},
                "country": {}
            }, predicates={
                "dest_city": "city",
                "dest_country": "country",
            }
        )

        self._when_compile_domain(dependencies={"?X.dest_country(X)": ["?X.dest_city(X)"]})

        self._then_result_has_field(
            "dependencies", {self._parse("?X.dest_country(X)"): set([self._parse("?X.dest_city(X)")])}
        )

    def test_default_questions(self):
        self._given_compiled_ontology(predicates={"price": "real"})
        self._when_compile_domain(default_questions=["?X.price(X)"])
        self._then_result_has_field("default_questions", [self._parse("?X.price(X)")])

    def test_default_questions_empty_by_default(self):
        self._given_compiled_ontology(predicates={"price": "real"})
        self._when_compile_domain()
        self._then_result_has_field("default_questions", [])

    def test_superactions(self):
        self._given_compiled_ontology(actions=set(["top", "buy"]))

        self._when_compile_domain(
            plans=[{
                "goal": "perform(top)",
                "plan": []
            }, {
                "goal": "perform(buy)",
                "plan": [],
                "superactions": ["top"]
            }]
        )

        self._then_result_has_field(
            "plans", [{
                "goal": self._parse("perform(top)"),
                "plan": Plan([])
            }, {
                "goal": self._parse("perform(buy)"),
                "plan": Plan([]),
                "superactions": [self._parse("top")]
            }]
        )

    def test_parameters(self):
        self._given_compiled_ontology(predicates={"price": "real"})

        self._when_compile_domain(parameters={"?X.price(X)": "{graphical_type=list}"})

        self._then_result_has_field("parameters", {self._parse("?X.price(X)"): {"graphical_type": "list"}})

    def test_predicate_parameters(self):
        self._given_compiled_ontology(sorts={"city": {}}, predicates={"price": "real", "dest_city": "city"})

        self._when_compile_domain(parameters={"price": "{background=[dest_city]}"})

        self._then_result_has_field("parameters", {self._parse("price"): {"background": [self._parse("dest_city")]}})


class GrammarCompilerTests(DddPyCompilerTestCase):
    def test_multiple_variants(self):
        self._given_compiled_ontology(actions=set(["buy"]))
        self._when_compile_grammar('buy = ["buy", "purchase"]')
        self._then_result_is(
            Node(
                Constants.GRAMMAR, {}, [
                    Node(
                        Constants.ACTION, {"name": "buy"}, [
                            Node(
                                Constants.ONE_OF, {},
                                [Node(Constants.ITEM, {}, ["buy"]),
                                 Node(Constants.ITEM, {}, ["purchase"])]
                            )
                        ]
                    )
                ]
            )
        )

    def test_action(self):
        self._given_compiled_ontology(actions=set(["buy"]))
        self._when_compile_grammar('buy = "purchase"')
        self._then_grammar_is([Node(Constants.ACTION, {"name": "buy"}, [Node(Constants.ITEM, {}, ["purchase"])])])

    def test_action_with_sortal_answer_of_custom_sort(self):
        self._given_compiled_ontology(actions=set(["buy"]), sorts={"city": {}})
        self._when_compile_grammar('buy = "purchase trip to <answer:city>"')
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "buy"},
                [Node(Constants.ITEM, {},
                      ["purchase trip to ", Node(Constants.SLOT, {"sort": "city"})])]
            )
        ])

    def test_action_with_sortal_string_answer(self):
        self._given_compiled_ontology(actions=set(["set_comment"]), predicates={"comment": "string"})
        self._when_compile_grammar('set_comment = "tell the hotel that <answer:string>"')
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "set_comment"},
                [Node(Constants.ITEM, {},
                      ["tell the hotel that ", Node(Constants.SLOT, {"sort": "string"})])]
            )
        ])

    def test_action_with_propositional_answer(self):
        self._given_compiled_ontology(actions=set(["buy"]), sorts={"city": {}}, predicates={"dest_city": "city"})
        self._when_compile_grammar('buy = "purchase trip to <answer:dest_city>"')
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "buy"},
                [Node(Constants.ITEM, {},
                      ["purchase trip to ", Node(Constants.SLOT, {"predicate": "dest_city"})])]
            )
        ])

    def test_predicate(self):
        self._given_compiled_ontology(predicates={"price": "real"})
        self._when_compile_grammar('price = "price information"')
        self._then_grammar_is([
            Node(Constants.PREDICATE, {"name": "price"}, [Node(Constants.ITEM, {}, ["price information"])])
        ])

    def test_user_question_with_sortal_answer_of_custom_sort(self):
        self._given_compiled_ontology(sorts={"city": {}}, predicates={"price": "real"})
        self._when_compile_grammar('price_user_question = "price for travelling to <answer:city>"')
        self._then_grammar_is([
            Node(
                Constants.USER_QUESTION, {"predicate": "price"},
                [Node(Constants.ITEM, {},
                      ["price for travelling to ", Node(Constants.SLOT, {"sort": "city"})])]
            )
        ])

    def test_user_question_with_propositional_answer(self):
        self._given_compiled_ontology(sorts={"city": {}}, predicates={"price": "real", "dest_city": "city"})
        self._when_compile_grammar('price_user_question = "price for travelling to <answer:dest_city>"')
        self._then_grammar_is([
            Node(
                Constants.USER_QUESTION, {"predicate": "price"}, [
                    Node(
                        Constants.ITEM, {},
                        ["price for travelling to ",
                         Node(Constants.SLOT, {"predicate": "dest_city"})]
                    )
                ]
            )
        ])

    def test_user_question_warning_for_undeclared_predicate(self):
        self._given_compiled_ontology()
        self._when_compile_grammar('undeclaredpredicate_user_question = "foo"')
        self._then_warning("failed to parse grammar entry key 'undeclaredpredicate_user_question'")

    def test_system_question(self):
        self._given_compiled_ontology(predicates={"price": "real"})
        self._when_compile_grammar('price_sys_question = "price information"')
        self._then_grammar_is([
            Node(Constants.SYS_QUESTION, {"predicate": "price"}, [Node(Constants.ITEM, {}, ["price information"])])
        ])

    def test_system_question_warning_for_undeclared_predicate(self):
        self._given_compiled_ontology()
        self._when_compile_grammar('undeclaredpredicate_sys_question = "foo"')
        self._then_warning("failed to parse grammar entry key 'undeclaredpredicate_sys_question'")

    def test_noun_phrase_as_single_variant(self):
        self._given_compiled_ontology()
        self._when_compile_grammar_with_noun_phrase_as_single_variant()
        self._then_result_has_expected_noun_phrase_as_single_variant()

    def _when_compile_grammar_with_noun_phrase_as_single_variant(self):
        self._compile_grammar("""
from tala.nl.gf.resource_eng import NP
top = NP("start view")
""")

    def _then_result_has_expected_noun_phrase_as_single_variant(self):
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "top"},
                [Node(Constants.NP, {}, [Node(Constants.INDEFINITE, {}, ["start view"])])]
            )
        ])

    def test_english_verb_phrase_as_single_variant(self):
        self._given_compiled_ontology(actions=set(["buy"]))
        self._when_compile_english_grammar_with_verb_phrase_as_single_variant()
        self._then_result_has_expected_english_verb_phrase_as_single_variant()

    def _when_compile_english_grammar_with_verb_phrase_as_single_variant(self):
        self._compile_grammar(
            """
from tala.nl.gf.resource_eng import VP
buy = VP("buy", "buy", "buying", "a ticket")
"""
        )

    def _then_result_has_expected_english_verb_phrase_as_single_variant(self):
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "buy"}, [
                    Node(
                        Constants.VP, {}, [
                            Node(Constants.INFINITIVE, {}, ["buy"]),
                            Node(Constants.IMPERATIVE, {}, ["buy"]),
                            Node(Constants.ING_FORM, {}, ["buying"]),
                            Node(Constants.OBJECT, {}, ["a ticket"])
                        ]
                    )
                ]
            )
        ])

    def test_swedish_verb_phrase_as_single_variant(self):
        self._given_compiled_ontology(actions=set(["buy"]))
        self._when_compile_swedish_grammar_with_verb_phrase_as_single_variant()
        self._then_result_has_expected_swedish_verb_phrase_as_single_variant()

    def _when_compile_swedish_grammar_with_verb_phrase_as_single_variant(self):
        self._compile_grammar(
            """
#-*- coding: utf-8 -*-
from tala.nl.gf.resource_sv import VP
buy = VP(u"köpa", u"köp", u"köper", u"en biljett")
"""
        )

    def _then_result_has_expected_swedish_verb_phrase_as_single_variant(self):
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "buy"}, [
                    Node(
                        Constants.VP, {}, [
                            Node(Constants.INFINITIVE, {}, [u"köpa"]),
                            Node(Constants.IMPERATIVE, {}, [u"köp"]),
                            Node(Constants.ING_FORM, {}, [u"köper"]),
                            Node(Constants.OBJECT, {}, [u"en biljett"])
                        ]
                    )
                ]
            )
        ])

    def test_english_verb_phrase_among_multiple_variants(self):
        self._given_compiled_ontology(actions=set(["buy"]))
        self._when_compile_english_grammar_with_verb_phrase_among_multiple_variants()
        self._then_result_has_expected_english_verb_phrase_among_multiple_variants()

    def _when_compile_english_grammar_with_verb_phrase_among_multiple_variants(self):
        self._compile_grammar(
            """
from tala.nl.gf.resource_eng import VP
buy = [VP("buy", "buy", "buying", "a ticket"), "purchase"]
"""
        )

    def _then_result_has_expected_english_verb_phrase_among_multiple_variants(self):
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "buy"}, [
                    Node(
                        Constants.ONE_OF, {}, [
                            Node(
                                Constants.ITEM, {}, [
                                    Node(
                                        Constants.VP, {}, [
                                            Node(Constants.INFINITIVE, {}, ["buy"]),
                                            Node(Constants.IMPERATIVE, {}, ["buy"]),
                                            Node(Constants.ING_FORM, {}, ["buying"]),
                                            Node(Constants.OBJECT, {}, ["a ticket"])
                                        ]
                                    )
                                ]
                            ),
                            Node(Constants.ITEM, {}, ["purchase"])
                        ]
                    )
                ]
            )
        ])

    def test_italian_noun_phrase_as_single_variant(self):
        self._given_compiled_ontology(actions=set(["settings"]))
        self._when_compile_italian_noun_phrase_as_single_variant()
        self._then_result_has_expected_italian_noun_phrase_as_single_variant()

    def _when_compile_italian_noun_phrase_as_single_variant(self):
        self._compile_grammar(
            """#-*- coding: utf-8 -*-
from tala.nl.gf.resource_it import NP, PLURAL, FEMININE
settings = NP(u"impostazioni", number=PLURAL, gender=FEMININE)
"""
        )

    def _then_result_has_expected_italian_noun_phrase_as_single_variant(self):
        self._then_grammar_is([
            Node(
                Constants.ACTION, {
                    "name": "settings",
                    "number": tala.nl.gf.resource.PLURAL,
                    "gender": tala.nl.gf.resource.FEMININE
                }, [Node(Constants.NP, {}, [Node(Constants.INDEFINITE, {}, [u"impostazioni"])])]
            )
        ])

    def test_italian_noun_phrase_among_multiple_variants(self):
        self._given_compiled_ontology(actions=set(["settings"]))
        self._when_compile_italian_noun_phrase_among_multiple_variants()
        self._then_result_has_expected_italian_noun_phrase_among_multiple_variants()

    def _when_compile_italian_noun_phrase_among_multiple_variants(self):
        self._compile_grammar(
            """#-*- coding: utf-8 -*-
from tala.nl.gf.resource_it import NP, PLURAL, FEMININE
settings = [NP(u"impostazioni", number=PLURAL, gender=FEMININE), u"vedere l'impostazioni"]
"""
        )

    def _then_result_has_expected_italian_noun_phrase_among_multiple_variants(self):
        self._then_grammar_is([
            Node(
                Constants.ACTION, {
                    "name": "settings",
                    "number": tala.nl.gf.resource.PLURAL,
                    "gender": tala.nl.gf.resource.FEMININE
                }, [
                    Node(
                        Constants.ONE_OF, {}, [
                            Node(
                                Constants.ITEM, {},
                                [Node(Constants.NP, {}, [Node(Constants.INDEFINITE, {}, [u"impostazioni"])])]
                            ),
                            Node(Constants.ITEM, {}, [u"vedere l'impostazioni"])
                        ]
                    )
                ]
            )
        ])

    def test_noun_phrase_among_multiple_variants(self):
        self._given_compiled_ontology()
        self._when_compile_grammar_with_noun_phrase_among_multiple_variants()
        self._then_result_has_expected_noun_phrase_among_multiple_variants()

    def _when_compile_grammar_with_noun_phrase_among_multiple_variants(self):
        self._compile_grammar(
            """
from tala.nl.gf.resource_eng import NP
top = [NP("start view", "the start view"), "main menu"]
"""
        )

    def _then_result_has_expected_noun_phrase_among_multiple_variants(self):
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "top"}, [
                    Node(
                        Constants.ONE_OF, {}, [
                            Node(
                                Constants.ITEM, {}, [
                                    Node(
                                        Constants.NP, {}, [
                                            Node(Constants.INDEFINITE, {}, ["start view"]),
                                            Node(Constants.DEFINITE, {}, ["the start view"])
                                        ]
                                    )
                                ]
                            ),
                            Node(Constants.ITEM, {}, ["main menu"])
                        ]
                    )
                ]
            )
        ])

    def test_answer_combination(self):
        self._given_compiled_ontology(sorts={"city": {}}, predicates={"dest_city": "city", "dept_city": "city"})
        self._when_compile_grammar(
            '''ANSWER = ["from <answer:dept_city> to <answer:dest_city>",
                                                "from <answer:dept_city>"]'''
        )
        self._then_grammar_is([
            Node(
                Constants.ANSWER_COMBINATION, {}, [
                    Node(
                        Constants.ONE_OF, {}, [
                            Node(
                                Constants.ITEM, {}, [
                                    "from ",
                                    Node(Constants.SLOT, {"predicate": "dept_city"}), " to ",
                                    Node(Constants.SLOT, {"predicate": "dest_city"})
                                ]
                            ),
                            Node(Constants.ITEM, {},
                                 ["from ", Node(Constants.SLOT, {"predicate": "dept_city"})])
                        ]
                    )
                ]
            )
        ])

    def test_report_started(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("IncomingCall")
        self._when_compile_grammar('IncomingCall_started = "incoming call"')
        self._then_grammar_is([
            Node(Constants.REPORT_STARTED, {"action": "IncomingCall"}, [Node(Constants.ITEM, {}, ["incoming call"])])
        ])

    def test_report_ended(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("IncomingCall")
        self._when_compile_grammar('IncomingCall_ended = "call ended"')
        self._then_grammar_is([
            Node(Constants.REPORT_ENDED, {"action": "IncomingCall"}, [Node(Constants.ITEM, {}, ["call ended"])])
        ])

    def test_action_with_empty_string(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("IncomingCall")
        self._when_compile_grammar('IncomingCall_ended = ""')
        self._then_grammar_is([
            Node(Constants.REPORT_ENDED, {"action": "IncomingCall"}, [Node(Constants.ITEM, {}, [])])
        ])

    def test_report_ended_warning_for_undeclared_service_action(self):
        self._given_compiled_ontology()
        self._given_service_interface_without_actions()
        self._when_compile_grammar('undeclared_action_ended = "string"')
        self._then_warning("failed to parse grammar entry key 'undeclared_action_ended'")

    def test_report_failed(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("CancelReservation")
        self._when_compile_grammar('CancelReservation_failed_no_reservation_exists = "there is no reservation"')
        self._then_grammar_is([
            Node(
                Constants.REPORT_FAILED, {
                    "action": "CancelReservation",
                    "reason": "no_reservation_exists"
                }, [Node(Constants.ITEM, {}, ["there is no reservation"])]
            )
        ])

    def test_individual(self):
        self._given_compiled_ontology(individuals={"city_1": "city"}, sorts={"city": {}})
        self._when_compile_grammar('city_1 = "paris"')
        self._then_grammar_is([Node(Constants.INDIVIDUAL, {"name": "city_1"}, [Node(Constants.ITEM, {}, ["paris"])])])

    def test_system_answer(self):
        self._given_compiled_ontology(predicates={"dest_city": "city"}, sorts={"city": {}})
        self._when_compile_grammar('dest_city_sys_answer = "to <individual>"')
        self._then_grammar_is([
            Node(
                Constants.SYS_ANSWER, {"predicate": "dest_city"},
                [Node(Constants.ITEM, {}, ["to ", Node(Constants.SLOT, {})])]
            )
        ])

    def test_system_answer_with_embedded_answer(self):
        self._given_compiled_ontology(predicates={"price": "real", "dest_city": "city"}, sorts={"city": {}})
        self._when_compile_grammar('price_sys_answer = "the price to <answer:dest_city> is <individual>"')
        self._then_grammar_is([
            Node(
                Constants.SYS_ANSWER, {"predicate": "price"}, [
                    Node(
                        Constants.ITEM, {}, [
                            "the price to ",
                            Node(Constants.SLOT, {"predicate": "dest_city"}), " is ",
                            Node(Constants.SLOT, {})
                        ]
                    )
                ]
            )
        ])

    def test_positive_system_answer(self):
        self._given_compiled_ontology(predicates={"qualified": "boolean"})
        self._when_compile_grammar('qualified_positive_sys_answer = "you are qualified"')
        self._then_grammar_is([
            Node(
                Constants.POSITIVE_SYS_ANSWER, {"predicate": "qualified"},
                [Node(Constants.ITEM, {}, ["you are qualified"])]
            )
        ])

    def test_negative_system_answer(self):
        self._given_compiled_ontology(predicates={"qualified": "boolean"})
        self._when_compile_grammar('qualified_negative_sys_answer = "you are not qualified"')
        self._then_grammar_is([
            Node(
                Constants.NEGATIVE_SYS_ANSWER, {"predicate": "qualified"},
                [Node(Constants.ITEM, {}, ["you are not qualified"])]
            )
        ])

    def test_action_title(self):
        self._given_compiled_ontology()
        self._when_compile_grammar('top_title = "main menu"')
        self._then_grammar_is([
            Node(Constants.ACTION_TITLE, {"action": "top"}, [Node(Constants.ITEM, {}, ["main menu"])])
        ])

    def test_issue_title(self):
        self._given_compiled_ontology(predicates={"price": "real"})
        self._when_compile_grammar('price_title = "price information"')
        self._then_grammar_is([
            Node(Constants.ISSUE_TITLE, {"predicate": "price"}, [Node(Constants.ITEM, {}, ["price information"])])
        ])

    def test_warning_for_invalid_title_key(self):
        self._given_compiled_ontology()
        self._when_compile_grammar('somethinginvalid_title = "foo"')
        self._then_warning("failed to parse grammar entry key 'somethinginvalid_title'")

    def test_validity(self):
        self._given_compiled_ontology(predicates={"dest_city": "city"}, sorts={"city": {}})
        self._given_service_interface_with_validity("CityValidity")
        self._when_compile_grammar('CityValidity = "invalid city <answer:dest_city>"')
        self._then_grammar_is([
            Node(
                Constants.VALIDITY, {"name": "CityValidity"},
                [Node(Constants.ITEM, {},
                      ["invalid city ", Node(Constants.SLOT, {"predicate": "dest_city"})])]
            )
        ])

    def test_prereport(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("MakeReservation")
        self._when_compile_grammar('MakeReservation_prereport = "making the reservation"')
        self._then_grammar_is([
            Node(
                Constants.PREREPORT, {"action": "MakeReservation"},
                [Node(Constants.ITEM, {}, ["making the reservation"])]
            )
        ])

    def test_preconfirm(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("MakeReservation")
        self._when_compile_grammar('MakeReservation_preconfirm = "would you like to make the reservation"')
        self._then_grammar_is([
            Node(
                Constants.PRECONFIRM, {"action": "MakeReservation"},
                [Node(Constants.ITEM, {}, ["would you like to make the reservation"])]
            )
        ])

    def test_greeting(self):
        self._given_compiled_ontology()
        self._when_compile_grammar('greeting = "Welcome"')
        self._then_grammar_is([Node(Constants.GREETING, {}, [Node(Constants.ITEM, {}, ["Welcome"])])])

    def test_invalid_key_is_causes_warning(self):
        self._given_compiled_ontology()
        self._when_compile_grammar('invalid_key = "string"')
        self._then_warning("failed to parse grammar entry key 'invalid_key'")

    def test_decompile_key_with_single_arg(self):
        self.assertEquals(
            "price_user_question",
            GrammarCompiler()._decompile_key(Node(Constants.USER_QUESTION, {"predicate": "price"}))
        )

    def test_decompile_key_with_multiple_args(self):
        parameters = OrderedDict()
        parameters["reason"] = "no_reservation_exists"
        parameters["action"] = "CancelReservation"
        self.assertEquals(
            "CancelReservation_failed_no_reservation_exists",
            GrammarCompiler()._decompile_key(Node(Constants.REPORT_FAILED, parameters))
        )

    def test_decompile_key_without_args(self):
        self.assertEquals("ANSWER", GrammarCompiler()._decompile_key(Node(Constants.ANSWER_COMBINATION, {})))

    def test_decompile_node(self):
        self.assertEquals(
            u'make_reservation = "make reservation"\n',
            GrammarCompiler().decompile_node(
                Node(Constants.ACTION, {'name': 'make_reservation'}, [u'make reservation'])
            )
        )



# encoding: utf-8

import os
import os.path
import shutil
import tempfile

import pytest
from pathlib import Path

from tala.utils.file_writer import UTF8FileWriter


class TestFileWriter(object):
    def setup(self):
        self._temp_dir = tempfile.mkdtemp(prefix=self.__class__.__name__)
        self._cwd = os.getcwd()
        os.chdir(self._temp_dir)

        self._file_writer = None

    def teardown(self):
        os.chdir(self._cwd)
        shutil.rmtree(self._temp_dir)

    @pytest.mark.parametrize("path", [
        "mocked-data.txt",
        Path("mocked-data.txt"),
    ])
    def test_write(self, path):
        self.given_file_writer_created_with(path)
        self.given_directories_created()
        self.when_writing("mocked data")
        self.then_file_contains(expected_file=Path("mocked-data.txt"), expected_contents="mocked data")

    def given_file_writer_created_with(self, path):
        self._file_writer = UTF8FileWriter(path)

    def given_directories_created(self):
        self._file_writer.create_directories()

    def when_writing(self, data):
        self._file_writer.write(data)

    def then_file_contains(self, expected_file, expected_contents):
        with expected_file.open() as actual_file:
            actual_contents = actual_file.read()
            assert actual_contents == expected_contents

    @pytest.mark.parametrize("path", [
        "mocked-data.txt",
        Path("mocked-data.txt"),
    ])
    def test_write_utf8(self, path):
        self.given_file_writer_created_with(path)
        self.given_directories_created()
        self.when_writing(u"småfisk")
        self.then_file_contains_utf8(expected_file=Path("mocked-data.txt"), expected_contents=u"småfisk")

    def then_file_contains_utf8(self, expected_file, expected_contents):
        with expected_file.open(encoding="utf-8") as actual_file:
            actual_contents = actual_file.read()
            assert actual_contents == expected_contents

    @pytest.mark.parametrize("path", [
        "mocked/data.txt",
        Path("mocked") / "data.txt",
    ])
    def test_write_in_nested_path(self, path):
        self.given_file_writer_created_with(path)
        self.given_directories_created()
        self.when_writing("mocked data")
        self.then_file_contains(expected_file=Path("mocked") / "data.txt", expected_contents="mocked data")



import os


class GrammarReader(object):
    @classmethod
    def read(cls, language_code, path=None):
        with open(cls.path(language_code, path)) as grammar_file:
            grammar_source = grammar_file.read()
        return grammar_source

    @staticmethod
    def path(language_code, path=None):
        def relative_path(language_code, path):
            filename = "grammar_%s.xml" % language_code
            if path:
                return os.path.join(path, filename)
            return filename

        path = relative_path(language_code, path)
        return os.path.abspath(path)

    @classmethod
    def xml_grammar_exists_for_language(cls, language_code, path=None):
        return os.path.exists(cls.path(language_code, path))



class EqualityAssertionTestCaseMixin:
    def assert_eq_returns_true_and_ne_returns_false_symmetrically(self, object1, object2):
        self.assertEquals(
            object1, object2, "%s of type %s != %s of type %s" %
            (unicode(object1), object1.__class__, unicode(object2), object2.__class__)
        )
        self.assertFalse(object1 != object2)

        self.assertEquals(
            object2, object1, "%s of type %s != %s of type %s" %
            (unicode(object2), object2.__class__, unicode(object1), object1.__class__)
        )
        self.assertFalse(object2 != object1)

    def assert_eq_returns_false_and_ne_returns_true_symmetrically(self, object1, object2):
        self.assertNotEquals(object1, object2)
        self.assertFalse(object1 == object2)

        self.assertNotEquals(object2, object1)
        self.assertFalse(object2 == object1)



import os
import re
import shutil
import subprocess
import tempfile

from pathlib import Path
import pytest

from tala.config import BackendConfig, DddConfig, RasaConfig
from tala.cli.console_formatting import BOLD, ENDC
from tala.cli import console_script
from tala.utils.chdir import chdir


class TempDirTestCase(object):
    def setup(self):
        self._temp_dir = tempfile.mkdtemp(prefix="TalaIntegrationTest")
        self._working_dir = os.getcwd()
        os.chdir(self._temp_dir)

    def teardown(self):
        os.chdir(self._working_dir)
        shutil.rmtree(self._temp_dir)


class ConsoleScriptTestCase(TempDirTestCase):
    def setup(self):
        super(ConsoleScriptTestCase, self).setup()
        self._process = None

    def _given_created_ddd_in_a_target_dir(self):
        self._create_ddd()

    def _create_ddd(self):
        self._run_tala_with(["create-ddd", "--target-dir", "test_root", "test_ddd"])

    def _given_changed_directory_to_target_dir(self):
        return chdir("test_root")

    def _given_changed_directory_to_ddd_folder(self):
        return chdir("test_root/test_ddd")

    def _then_result_is_successful(self):
        def assert_no_error_occured():
            pass

        assert_no_error_occured()

    def _when_running_command(self, command_line):
        self._stdout, self._stderr = self._run_command(command_line)

    def _run_tala_with(self, args):
        console_script.main(args)

    def _run_command(self, command_line):
        self._process = subprocess.Popen(command_line, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        return self._process.communicate()

    def _then_stderr_contains_constructive_error_message_for_missing_backend_config(self, config_path):
        pattern = "Expected backend config '.*{config}' to exist but it was not found. To create it, " "run 'tala create-backend-config --filename .*{config}'\.".format(
            config=config_path
        )
        assert re.search(pattern, self._stderr) is not None

    def _then_stderr_contains_constructive_error_message_for_missing_ddd_config(self, config_path):
        pattern = "Expected DDD config '.*{config}' to exist but it was not found. To create it, " "run 'tala create-ddd-config --filename .*{config}'\.".format(
            config=config_path
        )
        assert re.search(pattern, self._stderr) is not None

    def _given_config_overrides_missing_parent(self, path):
        self._replace_in_file(path, '"overrides": null', '"overrides": "missing_parent.json"')

    def _replace_in_file(self, path, old, new):
        with path.open() as f:
            old_contents = f.read()
        new_contents = old_contents.replace(old, new)
        with path.open("w") as f:
            f.write(new_contents)

    def _then_file_contains(self, filename, expected_string):
        actual_content = self._read_file(filename)
        assert expected_string in actual_content

    def _read_file(self, filename):
        with open(filename) as f:
            actual_content = f.read()
        return actual_content

    def _then_stderr_contains(self, string):
        assert string in self._stderr

    def _given_file_contains(self, filename, string):
        f = open(filename, "w")
        f.write(string)
        f.close()


class TestCreateDDD(ConsoleScriptTestCase):
    def test_create(self):
        self._when_creating_a_ddd()
        self._then_result_is_successful()

    def _when_creating_a_ddd(self):
        self._create_ddd()


class TestVersionIntegration(ConsoleScriptTestCase):
    def test_version(self):
        self._when_checking_tala_version()
        self._then_result_is_a_version_number()

    def _when_checking_tala_version(self):
        process = subprocess.Popen(["tala version"], stdout=subprocess.PIPE, shell=True)
        stdout, _ = process.communicate()
        self._result = stdout

    def _then_result_is_a_version_number(self):
        base_version = r"[0-9]+\.[0-9]+(\.[0-9]+)*"
        dev_version_suffix = r"\.dev[0-9]+\+[a-z0-9]+"
        dirty_version_suffix = r"\.d[0-9]{8}"

        release_version_regexp = base_version
        dev_version_regexp = base_version + dev_version_suffix
        dirty_version_regexp = base_version + dev_version_suffix + dirty_version_suffix

        is_version_regexp = r"(?:%s|%s|%s)" % (release_version_regexp, dev_version_regexp, dirty_version_regexp)

        assert re.search(is_version_regexp, self._result) is not None


class TestBackendConfigCreationIntegration(ConsoleScriptTestCase):
    def test_create_config_without_path_or_ddd(self):
        self._when_running_command("tala create-backend-config")
        self._then_backend_config_has_active_ddd(BackendConfig.default_name(), "")

    def test_create_backend_config_without_ddd(self):
        self._when_running_command("tala create-backend-config --filename test.config.json")
        self._then_backend_config_has_active_ddd("test.config.json", "")

    def _then_backend_config_has_active_ddd(self, config_path, expected_active_ddd):
        config = BackendConfig(config_path).read()
        assert expected_active_ddd == config["active_ddd"]

    def test_create_backend_config_with_ddd(self):
        self._when_running_command("tala create-backend-config --filename test.config.json --ddd test_ddd")
        self._then_backend_config_has_active_ddd("test.config.json", "test_ddd")


class TestConfigFileIntegration(ConsoleScriptTestCase):
    @pytest.mark.parametrize(
        "ConfigClass,command", [
            (BackendConfig, "create-backend-config"),
            (DddConfig, "create-ddd-config"),
            (RasaConfig, "create-rasa-config"),
        ]
    )
    def test_create_config_without_path(self, ConfigClass, command):
        self._when_running_command("tala {}".format(command))
        self._then_config_contains_defaults(ConfigClass, ConfigClass.default_name())

    def _then_config_contains_defaults(self, ConfigClass, name):
        expected_config = ConfigClass.default_config()
        actual_config = ConfigClass(name).read()
        assert expected_config == actual_config

    @pytest.mark.parametrize(
        "ConfigClass,command", [
            (BackendConfig, "create-backend-config"),
            (DddConfig, "create-ddd-config"),
            (RasaConfig, "create-rasa-config"),
        ]
    )
    def test_create_config_with_path(self, ConfigClass, command):
        self._when_running_command("tala {} --filename my_ddd.config.json".format(command))
        self._then_config_contains_defaults(ConfigClass, "my_ddd.config.json")

    @pytest.mark.parametrize(
        "name,command", [
            ("backend", "create-backend-config"),
            ("DDD", "create-ddd-config"),
            ("RASA", "create-rasa-config"),
        ]
    )
    def test_exception_raised_if_config_file_already_exists(self, name, command):
        self._given_config_was_created_with([command, "--filename", "test.config.json"])
        self._when_running_command("tala {} --filename test.config.json".format(command))
        self._then_stderr_contains(
            "Expected to be able to create {} config file 'test.config.json' but it already exists.".format(name)
        )

    def _given_config_was_created_with(self, arguments):
        self._run_tala_with(arguments)

    @pytest.mark.parametrize("command", [
        "create-backend-config",
        "create-ddd-config",
        "create-rasa-config",
    ])
    def test_config_file_not_overwritten(self, command):
        self._given_file_contains("test.config.json", "unmodified_mock_content")
        self._when_running_command("tala {} --filename test.config.json".format(command))
        self._then_file_contains("test.config.json", "unmodified_mock_content")

    @pytest.mark.parametrize("command", [
        "tala verify --config non_existing_config.json",
    ])
    def test_missing_config_causes_constructive_error_message(self, command):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_target_dir():
            self._when_running_command(command)
            self._then_stderr_contains_constructive_error_message_for_missing_backend_config("non_existing_config.json")

    @pytest.mark.parametrize("command", [
        "tala verify",
    ])
    def test_missing_parent_backend_config_causes_constructive_error_message(self, command):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_target_dir():
            self._given_config_overrides_missing_parent(Path(BackendConfig.default_name()))
            self._when_running_command(command)
            self._then_stderr_contains_constructive_error_message_for_missing_backend_config("missing_parent.json")

    @pytest.mark.parametrize("command", [
        "tala verify",
    ])
    def test_missing_parent_ddd_config_causes_constructive_error_message(self, command):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_ddd_folder():
            self._given_config_overrides_missing_parent(Path(DddConfig.default_name()))
        with self._given_changed_directory_to_target_dir():
            self._when_running_command(command)
            self._then_stderr_contains_constructive_error_message_for_missing_ddd_config("missing_parent.json")


class TestVerifyIntegration(ConsoleScriptTestCase):
    def setup(self):
        super(TestVerifyIntegration, self).setup()

    def test_that_verifying_boilerplate_ddd_succeeds(self):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_target_dir():
            self._when_verifying()
        self._then_result_is_successful()

    def _when_verifying(self):
        self._run_tala_with(["verify"])

    def test_stdout_when_verifying_boilerplate_ddd(self):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_target_dir():
            self._when_running_command("tala verify")
        self._then_stdout_matches(
            "^Verifying models for DDD 'test_ddd'.\n"
            "\[eng\] Verifying grammar.\n"
            "\[eng\] Finished verifying grammar.\n"
            "\[eng\] RASA NLU is disabled, skipping its grammar verification.\n"
            "Finished verifying models for DDD 'test_ddd'.\n$"
        )

    def _then_stdout_matches(self, expected_pattern):
        assert re.match(expected_pattern,
                        self._stdout) is not None, "Expected stdout to match '{}' but got '{}'".format(
                            expected_pattern, self._stdout
                        )

    def test_stdout_when_verifying_boilerplate_ddd_with_rasa_enabled(self):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_ddd_folder():
            self._given_enabled_rasa()
        with self._given_changed_directory_to_target_dir():
            self._given_rasa_config_created()
            self._when_running_command("tala verify")
        self._then_stdout_matches(
            "^Verifying models for DDD 'test_ddd'.\n"
            "\[eng\] Verifying grammar.\n"
            "\[eng\] Finished verifying grammar.\n"
            "\[eng\] Verifying grammar for RASA NLU.\n"
            "\[eng\] Finished verifying grammar for RASA NLU.\n"
            "Finished verifying models for DDD 'test_ddd'.\n$"
        )

    def _given_rasa_config_created(self):
        self._run_tala_with(["create-rasa-config"])

    def _given_enabled_rasa(self):
        self._replace_in_file(Path(DddConfig.default_name()), '"enable_rasa_nlu": false', '"enable_rasa_nlu": true')

    def test_stderr_when_verifying_boilerplate_ddd(self):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_target_dir():
            self._when_running_command("tala verify")
        self._then_stderr_is_empty()

    def _then_stderr_is_empty(self):
        assert self._stderr == ""

    def test_stderr_when_verifying_boilerplate_ddd_with_rasa_enabled(self):
        def format_regular_expression_as_warning(regular_expression):
            return "{}{}{}".format(re.escape(BOLD), regular_expression, re.escape(ENDC))

        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_ddd_folder():
            self._given_enabled_rasa()
        with self._given_changed_directory_to_target_dir():
            self._given_rasa_config_created()
            self._when_running_command("tala verify")
        self._then_stderr_matches_only(
            format_regular_expression_as_warning(
                "UserWarning: The support for RASA NLU is still in BETA. Talk gently. "
                "It is currently enabled in DDD 'test_ddd'.\n"
            )
        )

    def _then_stderr_matches_only(self, pattern):
        expected_pattern = "^{}$".format(pattern)
        match = re.match(expected_pattern, self._stderr)
        assert match is not None, "Expected pattern {!r} to match stderr {!r} but it didn't".format(
            expected_pattern, self._stderr
        )

    def test_verify_creates_no_build_folders(self):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_target_dir():
            self._when_running_command("tala verify")
        self._then_there_are_no_build_folders()

    def _then_there_are_no_build_folders(self):
        ddd_folder = Path("test_root") / "test_ddd" / "grammar"
        build_folders = [path for path in ddd_folder.iterdir() if path.is_dir() and path.name.startswith("build")]
        assert not any(build_folders), "Expected no build folders but got {}".format(build_folders)

    def test_verify_creates_no_build_folders_with_rasa_enabled(self):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_ddd_folder():
            self._given_enabled_rasa()
        with self._given_changed_directory_to_target_dir():
            self._given_rasa_config_created()
            self._when_running_command("tala verify")
        self._then_there_are_no_build_folders()

    def test_verify_returncode_with_schema_violation(self):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_ddd_folder():
            self._given_schema_violation_in_ontology()
        with self._given_changed_directory_to_target_dir():
            self._when_running_command("tala verify")
        self._then_returncode_signals_error()

    def _given_schema_violation_in_ontology(self):
        self._replace_in_file(Path("ontology.xml"), "ontology", "hello")

    def _then_returncode_signals_error(self):
        assert self._process.returncode != 0

    def test_verify_stderr_with_schema_violation(self):
        self._given_created_ddd_in_a_target_dir()
        with self._given_changed_directory_to_ddd_folder():
            self._given_schema_violation_in_ontology()
        with self._given_changed_directory_to_target_dir():
            self._when_running_command("tala verify")
        self._then_stderr_contains(
            "Expected ontology.xml compliant with schema but it's in violation: "
            "Element 'hello': "
            "No matching global declaration available for the validation root., line 2"
        )
